"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bcrypt_1 = __importDefault(require("bcrypt"));
const mongodb_1 = require("mongodb");
const schema_1 = require("./schema");
const utility_1 = require("../utility");
class MongoDBUserAdapter {
    constructor(db, bcryptHashCostFactor) {
        this.users = db.collection(schema_1.CollectionName.Users);
        this.bcryptHashCostFactor = bcryptHashCostFactor;
    }
    async createUser({ email, name, password, roleIDs }) {
        try {
            const passwordHash = await bcrypt_1.default.hash(password, this.bcryptHashCostFactor);
            const { insertedId: id } = await this.users.insertOne({
                createdAt: new Date(),
                modifiedAt: new Date(),
                email,
                name,
                roleIDs,
                password: passwordHash
            });
            return { id, email, name, roleIDs };
        }
        catch (err) {
            if (err instanceof mongodb_1.MongoError && err.code === utility_1.MongoErrorCode.DuplicateKey) {
                throw new Error('Email address already exists!');
            }
            throw err;
        }
    }
    async getUser(email) {
        const user = await this.users.findOne({ email });
        if (!user)
            return null;
        return {
            id: user._id,
            email: user.email,
            name: user.name,
            roleIDs: user.roleIDs
        };
    }
    async getUsersByID(ids) {
        const users = await this.users.find({ _id: { $in: ids } }).toArray();
        return users.map(user => ({
            id: user._id,
            email: user.email,
            name: user.name,
            roleIDs: user.roleIDs
        }));
    }
    async getUserForCredentials({ email, password }) {
        const user = await this.users.findOne({ email });
        if (!user)
            return null;
        const passwordMatches = await bcrypt_1.default.compare(password, user.password);
        if (!passwordMatches)
            return null;
        return {
            id: user._id,
            email: user.email,
            name: user.name,
            roleIDs: user.roleIDs
        };
    }
    async getUserByID(id) {
        const user = await this.users.findOne({ _id: id });
        if (!user)
            return null;
        return {
            id: user._id,
            email: user.email,
            name: user.name,
            roleIDs: user.roleIDs
        };
    }
}
exports.MongoDBUserAdapter = MongoDBUserAdapter;
//# sourceMappingURL=user.js.map