"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const schema_1 = require("./db/schema");
const SessionDocumentTTL = 60 * 60 * 24; // 24h
exports.Migrations = [
    {
        version: 0,
        async migrate(db, locale) {
            const migrations = await db.createCollection(schema_1.CollectionName.Migrations, { strict: true });
            await migrations.createIndex({ name: 1 }, { unique: true });
            const users = await db.createCollection(schema_1.CollectionName.Users, {
                strict: true
            });
            await users.createIndex({ email: 1 }, { unique: true });
            const sessions = await db.createCollection(schema_1.CollectionName.Sessions, {
                strict: true
            });
            await sessions.createIndex({ userID: 1 });
            await sessions.createIndex({ token: 1 }, { unique: true });
            await sessions.createIndex({ expiresAt: 1 }, { expireAfterSeconds: SessionDocumentTTL });
            const navigations = await db.createCollection(schema_1.CollectionName.Navigations, { strict: true });
            await navigations.createIndex({ createdAt: -1 });
            await navigations.createIndex({ modifiedAt: -1 });
            await navigations.createIndex({ name: 1 });
            await navigations.createIndex({ key: 1 }, { unique: true });
            const authors = await db.createCollection(schema_1.CollectionName.Authors, { strict: true });
            await authors.createIndex({ createdAt: -1 });
            await authors.createIndex({ modifiedAt: -1 });
            await authors.createIndex({ name: 1 });
            await authors.createIndex({ slug: 1 }, { unique: true });
            const images = await db.createCollection(schema_1.CollectionName.Images, { strict: true });
            await images.createIndex({ createdAt: -1 });
            await images.createIndex({ modifiedAt: -1 });
            await images.createIndex({ title: 1 });
            await images.createIndex({ tags: 1 }, { collation: { locale, strength: 2 } });
            const articles = await db.createCollection(schema_1.CollectionName.Articles, {
                strict: true
            });
            await articles.createIndex({ createdAt: -1 });
            await articles.createIndex({ modifiedAt: -1 });
            await articles.createIndex({ 'published.publishedAt': -1 });
            await articles.createIndex({ 'published.updatedAt': -1 });
            await articles.createIndex({ 'pending.publishAt': -1 });
            await articles.createIndex({ 'draft.tags': 1 }, { collation: { locale, strength: 2 } });
            await articles.createIndex({ 'pending.tags': 1 }, { collation: { locale, strength: 2 } });
            await articles.createIndex({ 'published.tags': 1 }, { collation: { locale, strength: 2 } });
            await db.createCollection(schema_1.CollectionName.ArticlesHistory, {
                strict: true
            });
            const pages = await db.createCollection(schema_1.CollectionName.Pages, {
                strict: true
            });
            await pages.createIndex({ createdAt: -1 });
            await pages.createIndex({ modifiedAt: -1 });
            await pages.createIndex({ 'published.publishedAt': -1 });
            await pages.createIndex({ 'published.updatedAt': -1 });
            await pages.createIndex({ 'pending.publishAt': -1 });
            await pages.createIndex({ 'draft.tags': 1 }, { collation: { locale, strength: 2 } });
            await pages.createIndex({ 'pending.tags': 1 }, { collation: { locale, strength: 2 } });
            await pages.createIndex({ 'published.tags': 1 }, { collation: { locale, strength: 2 } });
            // TODO: Add unique index for slug on published page
            await db.createCollection(schema_1.CollectionName.PagesHistory, {
                strict: true
            });
        }
    },
    {
        //  Fix incorrect migration index. Add user roles, Add name and roleIDs to users.
        version: 1,
        async migrate(db, locale) {
            const migrations = db.collection(schema_1.CollectionName.Migrations);
            await migrations.dropIndex('name_1');
            await migrations.createIndex({ version: 1 }, { unique: true });
            const userRoles = await db.createCollection(schema_1.CollectionName.UserRoles, {
                strict: true
            });
            await userRoles.createIndex({ name: 1 }, { unique: true });
            userRoles.insertMany([
                {
                    _id: 'admin',
                    createdAt: new Date(),
                    modifiedAt: new Date(),
                    systemRole: true,
                    name: 'Admin',
                    description: 'Administrator Role',
                    permissionIDs: []
                },
                {
                    _id: 'editor',
                    createdAt: new Date(),
                    modifiedAt: new Date(),
                    systemRole: true,
                    name: 'Editor',
                    description: 'Editor Role',
                    permissionIDs: []
                }
            ]);
            const user = db.collection(schema_1.CollectionName.Users);
            user.updateMany({}, [
                {
                    $set: {
                        name: '$email',
                        roleIDs: ['admin']
                    }
                }
            ]);
        }
    },
    {
        // Add peering and token collections and migrate ArticleTeaserGridBlock to TeaserGridBlock.
        version: 2,
        async migrate(db) {
            const userRoles = db.collection(schema_1.CollectionName.UserRoles);
            userRoles.insertOne({
                _id: 'peer',
                createdAt: new Date(),
                modifiedAt: new Date(),
                systemRole: true,
                name: 'Peer',
                description: 'Peer Role',
                permissionIDs: []
            });
            await db.createCollection(schema_1.CollectionName.PeerProfiles, { strict: true });
            const peers = await db.createCollection(schema_1.CollectionName.Peers, { strict: true });
            await peers.createIndex({ slug: 1 }, { unique: true });
            const tokens = await db.createCollection(schema_1.CollectionName.Tokens, {
                strict: true
            });
            await tokens.createIndex({ name: 1 }, { unique: true });
            const filter = {
                $or: [
                    { 'draft.blocks.type': 'articleTeaserGrid' },
                    { 'published.blocks.type': 'articleTeaserGrid' },
                    { 'pending.blocks.type': 'articleTeaserGrid' }
                ]
            };
            function mapArticleTeaserGridBlock(block) {
                if (block.type === 'articleTeaserGrid') {
                    return {
                        type: 'teaserGrid',
                        teasers: block.teasers.map((teaser) => teaser
                            ? {
                                type: 'article',
                                style: 'default',
                                articleID: teaser.articleID
                            }
                            : null),
                        numColumns: block.numColumns
                    };
                }
                return block;
            }
            const articles = db.collection(schema_1.CollectionName.Articles);
            const migrationArticles = await articles.find(filter).toArray();
            for (const article of migrationArticles) {
                if (article.draft) {
                    article.draft.blocks = article.draft.blocks.map(mapArticleTeaserGridBlock);
                }
                if (article.pending) {
                    article.pending.blocks = article.pending.blocks.map(mapArticleTeaserGridBlock);
                }
                if (article.published) {
                    article.published.blocks = article.published.blocks.map(mapArticleTeaserGridBlock);
                }
                await articles.findOneAndReplace({ _id: article._id }, article);
            }
            const pages = db.collection(schema_1.CollectionName.Pages);
            const migrationPages = await pages.find(filter).toArray();
            for (const page of migrationPages) {
                if (page.draft) {
                    page.draft.blocks = page.draft.blocks.map(mapArticleTeaserGridBlock);
                }
                if (page.pending) {
                    page.pending.blocks = page.pending.blocks.map(mapArticleTeaserGridBlock);
                }
                if (page.published) {
                    page.published.blocks = page.published.blocks.map(mapArticleTeaserGridBlock);
                }
                await pages.findOneAndReplace({ _id: page._id }, page);
            }
        }
    }
];
exports.LatestMigration = exports.Migrations[exports.Migrations.length - 1];
//# sourceMappingURL=migration.js.map